@@ -39,10 +39,11 @@
 #include "udiskslinuxblockobject.h"
 #include "udiskslinuxfsinfo.h"
 #include "udisksdaemon.h"
-#include "udiskscleanup.h"
+#include "udisksstate.h"
 #include "udisksdaemonutil.h"
 #include "udisksmountmonitor.h"
 #include "udisksmount.h"
+#include "udiskslinuxdevice.h"
 
 /**
  * SECTION:udiskslinuxfilesystem
@@ -135,7 +136,7 @@ udisks_linux_filesystem_update (UDisksLinuxFilesystem  *filesystem,
                                 UDisksLinuxBlockObject *object)
 {
   UDisksMountMonitor *mount_monitor;
-  GUdevDevice *device;
+  UDisksLinuxDevice *device;
   GPtrArray *p;
   GList *mounts;
   GList *l;
@@ -144,7 +145,7 @@ udisks_linux_filesystem_update (UDisksLinuxFilesystem  *filesystem,
   device = udisks_linux_block_object_get_device (object);
 
   p = g_ptr_array_new ();
-  mounts = udisks_mount_monitor_get_mounts_for_dev (mount_monitor, g_udev_device_get_device_number (device));
+  mounts = udisks_mount_monitor_get_mounts_for_dev (mount_monitor, g_udev_device_get_device_number (device->udev_device));
   /* we are guaranteed that the list is sorted so if there are
    * multiple mounts we'll always get the same order
    */
@@ -274,7 +275,7 @@ typedef struct
 
 /* ---------------------- vfat -------------------- */
 
-static const gchar *vfat_defaults[] = { "uid=", "gid=", "shortname=mixed", "dmask=0077", "utf8=1", "showexec", NULL };
+static const gchar *vfat_defaults[] = { "uid=", "gid=", "shortname=mixed", "dmask=0077", "utf8=1", "showexec", "flush", NULL };
 static const gchar *vfat_allow[] = { "flush", "utf8=", "shortname=", "umask=", "dmask=", "fmask=", "codepage=", "iocharset=", "usefree", "showexec", NULL };
 static const gchar *vfat_allow_uid_self[] = { "uid=", NULL };
 static const gchar *vfat_allow_gid_self[] = { "gid=", NULL };
@@ -840,6 +841,8 @@ calculate_mount_point (UDisksDaemon              *daemon,
                        const gchar               *fs_type,
                        GError                   **error)
 {
+  UDisksLinuxBlockObject *object = NULL;
+  gboolean fs_shared = FALSE;
   const gchar *label = NULL;
   const gchar *uuid = NULL;
   gchar *escaped_user_name = NULL;
@@ -858,10 +861,25 @@ calculate_mount_point (UDisksDaemon              *daemon,
       uuid = udisks_block_get_id_uuid (block);
     }
 
+  object = udisks_daemon_util_dup_object (block, NULL);
+  if (object != NULL)
+    {
+      UDisksLinuxDevice *device = udisks_linux_block_object_get_device (object);
+      if (device != NULL)
+        {
+          if (device->udev_device != NULL)
+            {
+              /* TODO: maybe introduce Block:HintFilesystemShared instead of pulling it directly from the udev device */
+              fs_shared = g_udev_device_get_property_as_boolean (device->udev_device, "UDISKS_FILESYSTEM_SHARED");
+            }
+          g_object_unref (device);
+        }
+    }
+
   /* If we know the user-name and it doesn't have any '/' character in
    * it, mount in /run/media/$USER
    */
-  if (user_name != NULL && strstr (user_name, "/") == NULL)
+  if (!fs_shared && (user_name != NULL && strstr (user_name, "/") == NULL))
     {
       mount_dir = g_strdup_printf ("/run/media/%s", user_name);
       if (!g_file_test (mount_dir, G_FILE_TEST_EXISTS))
@@ -962,6 +980,7 @@ calculate_mount_point (UDisksDaemon              *daemon,
   g_free (mount_dir);
 
  out:
+  g_clear_object (&object);
   g_free (escaped_user_name);
   return mount_point;
 }
@@ -1102,10 +1121,10 @@ handle_mount (UDisksFilesystem       *filesystem,
   UDisksObject *object;
   UDisksBlock *block;
   UDisksDaemon *daemon;
-  UDisksCleanup *cleanup;
-  pid_t caller_pid;
+  UDisksState *state;
   uid_t caller_uid;
   gid_t caller_gid;
+  pid_t caller_pid;
   const gchar * const *existing_mount_points;
   const gchar *probed_fs_usage;
   gchar *fs_type_to_use;
@@ -1148,7 +1167,7 @@ handle_mount (UDisksFilesystem       *filesystem,
 
   block = udisks_object_peek_block (object);
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
-  cleanup = udisks_daemon_get_cleanup (daemon);
+  state = udisks_daemon_get_state (daemon);
 
   /* check if mount point is managed by e.g. /etc/fstab or similar */
   if (is_system_managed (block, &mount_point_to_use, &fstab_mount_options))
@@ -1260,6 +1279,7 @@ handle_mount (UDisksFilesystem       *filesystem,
     mount_fstab_again:
       if (!udisks_daemon_launch_spawned_job_sync (daemon,
                                                   object,
+                                                  "filesystem-mount", caller_uid,
                                                   NULL,  /* GCancellable */
                                                   mount_fstab_as_root ? 0 : caller_uid, /* uid_t run_as_uid */
                                                   mount_fstab_as_root ? 0 : caller_uid, /* uid_t run_as_euid */
@@ -1308,11 +1328,11 @@ handle_mount (UDisksFilesystem       *filesystem,
                      caller_uid);
 
       /* update the mounted-fs file */
-      udisks_cleanup_add_mounted_fs (cleanup,
-                                     mount_point_to_use,
-                                     udisks_block_get_device_number (block),
-                                     caller_uid,
-                                     TRUE); /* fstab_mounted */
+      udisks_state_add_mounted_fs (state,
+                                   mount_point_to_use,
+                                   udisks_block_get_device_number (block),
+                                   caller_uid,
+                                   TRUE); /* fstab_mounted */
 
       udisks_filesystem_complete_mount (filesystem, invocation, mount_point_to_use);
       goto out;
@@ -1436,6 +1456,7 @@ handle_mount (UDisksFilesystem       *filesystem,
   /* run mount(8) */
   if (!udisks_daemon_launch_spawned_job_sync (daemon,
                                               object,
+                                              "filesystem-mount", caller_uid,
                                               NULL, /* GCancellable */
                                               0,    /* uid_t run_as_uid */
                                               0,    /* uid_t run_as_euid */
@@ -1462,11 +1483,11 @@ handle_mount (UDisksFilesystem       *filesystem,
     }
 
   /* update the mounted-fs file */
-  udisks_cleanup_add_mounted_fs (cleanup,
-                                 mount_point_to_use,
-                                 udisks_block_get_device_number (block),
-                                 caller_uid,
-                                 FALSE); /* fstab_mounted */
+  udisks_state_add_mounted_fs (state,
+                               mount_point_to_use,
+                               udisks_block_get_device_number (block),
+                               caller_uid,
+                               FALSE); /* fstab_mounted */
 
   udisks_notice ("Mounted %s at %s on behalf of uid %d",
                  udisks_block_get_device (block),
@@ -1516,7 +1537,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
   UDisksObject *object;
   UDisksBlock *block;
   UDisksDaemon *daemon;
-  UDisksCleanup *cleanup;
+  UDisksState *state;
   gchar *mount_point;
   gchar *fstab_mount_options;
   gchar *escaped_mount_point;
@@ -1551,7 +1572,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
 
   block = udisks_object_peek_block (object);
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
-  cleanup = udisks_daemon_get_cleanup (daemon);
+  state = udisks_daemon_get_state (daemon);
   system_managed = FALSE;
 
   if (options != NULL)
@@ -1601,6 +1622,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
       /* right now -l is the only way to "force unmount" file systems... */
       if (!udisks_daemon_launch_spawned_job_sync (daemon,
                                                   object,
+                                                  "filesystem-unmount", caller_uid,
                                                   NULL, /* GCancellable */
                                                   unmount_fstab_as_root ? 0 : caller_uid, /* uid_t run_as_uid */
                                                   unmount_fstab_as_root ? 0 : caller_uid, /* uid_t run_as_euid */
@@ -1656,10 +1678,10 @@ handle_unmount (UDisksFilesystem       *filesystem,
     }
 
   error = NULL;
-  mount_point = udisks_cleanup_find_mounted_fs (cleanup,
-                                                udisks_block_get_device_number (block),
-                                                &mounted_by_uid,
-                                                &fstab_mounted);
+  mount_point = udisks_state_find_mounted_fs (state,
+                                              udisks_block_get_device_number (block),
+                                              &mounted_by_uid,
+                                              &fstab_mounted);
   if (mount_point == NULL)
     {
       /* allow unmounting stuff not mentioned in mounted-fs, but treat it like root mounted it */
@@ -1698,6 +1720,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
       escaped_mount_point = udisks_daemon_util_escape_and_quote (mount_point);
       rc = udisks_daemon_launch_spawned_job_sync (daemon,
                                                   object,
+                                                  "filesystem-unmount", caller_uid,
                                                   NULL, /* GCancellable */
                                                   0,    /* uid_t run_as_uid */
                                                   0,    /* uid_t run_as_euid */
@@ -1713,6 +1736,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
       /* mount_point == NULL */
       rc = udisks_daemon_launch_spawned_job_sync (daemon,
                                                   object,
+                                                  "filesystem-unmount", caller_uid,
                                                   NULL, /* GCancellable */
                                                   0,    /* uid_t run_as_uid */
                                                   0,    /* uid_t run_as_euid */
@@ -1735,7 +1759,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
       goto out;
     }
 
-  /* OK, filesystem unmounted.. the cleanup routines will remove the mountpoint for us */
+  /* OK, filesystem unmounted.. the state/cleanup routines will remove the mountpoint for us */
 
   udisks_notice ("Unmounted %s on behalf of uid %d",
                  udisks_block_get_device (block),
@@ -1795,6 +1819,8 @@ handle_set_label (UDisksFilesystem       *filesystem,
   UDisksBaseJob *job;
   const gchar *action_id;
   const gchar *message;
+  uid_t caller_uid;
+  gid_t caller_gid;
   pid_t caller_pid;
   gchar *command;
   gchar *tmp;
@@ -1827,6 +1853,20 @@ handle_set_label (UDisksFilesystem       *filesystem,
       goto out;
     }
 
+  error = NULL;
+  if (!udisks_daemon_util_get_caller_uid_sync (daemon,
+                                               invocation,
+                                               NULL /* GCancellable */,
+                                               &caller_uid,
+                                               &caller_gid,
+                                               NULL,
+                                               &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      g_error_free (error);
+      goto out;
+    }
+
   probed_fs_usage = udisks_block_get_id_usage (block);
   probed_fs_type = udisks_block_get_id_type (block);
 
@@ -1857,15 +1897,17 @@ handle_set_label (UDisksFilesystem       *filesystem,
    * question in this case, check in advance */
   if (g_strcmp0 (probed_fs_type, "vfat") == 0)
     {
-      for (tmp = "\"*/:<>?\\|"; *tmp; ++tmp)
+      const gchar *forbidden = "\"*/:<>?\\|";
+      guint n;
+      for (n = 0; forbidden[n] != 0; n++)
         {
-          if (strchr (label, *tmp) != NULL)
+          if (strchr (label, forbidden[n]) != NULL)
             {
               g_dbus_method_invocation_return_error (invocation,
                                                      UDISKS_ERROR,
                                                      UDISKS_ERROR_NOT_SUPPORTED,
                                                      "character '%c' not supported in VFAT labels",
-                                                     *tmp);
+                                                     forbidden[n]);
                goto out;
             }
         }
@@ -1898,13 +1940,16 @@ handle_set_label (UDisksFilesystem       *filesystem,
    * will be replaced by the name of the drive/device in question
    */
   message = N_("Authentication is required to change the filesystem label on $(drive)");
-  if (udisks_block_get_hint_system (block))
+  if (!udisks_daemon_util_setup_by_user (daemon, object, caller_uid))
     {
-      action_id = "org.freedesktop.udisks2.modify-device-system";
-    }
-  else if (!udisks_daemon_util_on_same_seat (daemon, UDISKS_OBJECT (object), caller_pid))
-    {
-      action_id = "org.freedesktop.udisks2.modify-device-other-seat";
+      if (udisks_block_get_hint_system (block))
+        {
+          action_id = "org.freedesktop.udisks2.modify-device-system";
+        }
+      else if (!udisks_daemon_util_on_same_seat (daemon, UDISKS_OBJECT (object), caller_pid))
+        {
+          action_id = "org.freedesktop.udisks2.modify-device-other-seat";
+        }
     }
 
   /* Check that the user is actually authorized to change the
@@ -1931,6 +1976,7 @@ handle_set_label (UDisksFilesystem       *filesystem,
 
   job = udisks_daemon_launch_spawned_job (daemon,
                                           object,
+                                          "filesystem-modify", caller_uid,
                                           NULL, /* cancellable */
                                           0,    /* uid_t run_as_uid */
                                           0,    /* uid_t run_as_euid */
